#include <pthread.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <curl/curl.h>
#include <stdio.h>
#include <sys/file.h>
#include "fmt.h"
#include "tags/include/tags.h"
#include "queue.h"
#include "mixi_music.h"
#include "config.h"
#include <glib.h>

#define MIXI_LOGIN_URL "https://upload.mixi.jp/music/login.pl"
#define MIXI_USER_AGENT "mmm/v0.1 (by glucose)"

#define MIXI_RETRY_SHORT 60
#define MIXI_RETRY_LONG(i) (60 << (i))
#define MIXI_RETRY_MAX 7200
#define MIXI_RETRY_SHORT_COUNT 5
#define MIXI_RETRY_LONG_COUNT 12

#define CACHE_SIZE 1024

static int mixi_hs_status;
static int mixi_hs_timeout;
static int mixi_hs_errors; 
static int mixi_sb_errors;
static int mixi_bad_users;
static int mixi_submit_timeout;
static int mixi_srv_res_size;
static int mixi_giveup;
static int mixi_major_error_present;

static char *mixi_upload_url;
static char *mixi_email;
static char *mixi_password;
static char *mixi_session_id;
static char *mixi_srv_res;
static char mixi_curl_errbuf[CURL_ERROR_SIZE];
static char *mixi_major_error;

static void dump_queue();

/* Error functions */

static void mixi_throw_error(char *errortxt)
{
  mixi_major_error_present = 1;

  if(mixi_major_error == NULL)
	mixi_major_error = strdup(errortxt);

  return;
}

int
mixi_catch_error(void)
{
  return mixi_major_error_present;
}

char *
mixi_fetch_error(void)
{
  return mixi_major_error;
}

void
mixi_clear_error(void)
{
  mixi_major_error_present = 0;
  if(mixi_major_error != NULL)
	free(mixi_major_error);

  mixi_major_error = NULL;

  return;
}


static size_t
mixi_store_res(void *ptr, size_t size, size_t nmemb, void *stream)
{
  int len = size * nmemb;

  mixi_srv_res = realloc(mixi_srv_res, mixi_srv_res_size + len + 1);

  memcpy(mixi_srv_res + mixi_srv_res_size, ptr, len);

  mixi_srv_res_size += len;

  return len;
}

static void
mixi_free_res(void)
{
  if(mixi_srv_res != NULL) free(mixi_srv_res);

  mixi_srv_res = NULL;
  mixi_srv_res_size = 0;
}


static int
mixi_parse_hs_res(void)
{
  char *ptr1, *ptr2;

  if (!mixi_srv_res_size) {
	pdebug("No reply from server", DEBUG);
	return -1;
  }

  *(mixi_srv_res + mixi_srv_res_size) = '\0';

  ptr1 = mixi_srv_res;

  if (strncmp(ptr1, "status=", 7) == 0) {
	ptr1 += 7;	
	if ((ptr2 = strchr(ptr1, '\n'))) *ptr2 = '\0';
  } else {
	mixi_throw_error("Bad server message.");
	return -1;
  }

  if (strcmp(ptr1, "OK") == 0) {

	ptr1 = ptr2 + 1;

	if (strncmp(ptr1, "session_id=", 11) == 0) {
	  ptr1 += 11;
	  if ((ptr2 = strchr(ptr1, '\n'))) *ptr2 = '\0';
	} else {
	  mixi_throw_error("Session ID not found.");
	  return -1;
	}

	if (mixi_session_id) free(mixi_session_id);
	mixi_session_id = strdup(ptr1);
	
	ptr1 = ptr2 + 1;

	if (strncmp(ptr1, "url=", 4) == 0) {
	  ptr1 += 4;
	  if ((ptr2 = strchr(ptr1, '\n'))) *ptr2 = '\0';
	} else {
	  mixi_throw_error("Upload URL not found.");
	  return -1;
	}

	pdebug(mixi_fmt_vastr("session ID: %s", mixi_session_id), DEBUG);

	mixi_upload_url = strdup(ptr1);

	return 0;

  } else if (strcmp(ptr1, "BAD_PASSWORD") == 0) {

	mixi_throw_error("Incorrect E-Mail/Password.\n"
					 "Please fix in configuration.");

	pdebug("incorrect email/password", DEBUG);

	return -1;

  } else {
	mixi_throw_error("Unkown server-reply.");
	pdebug(mixi_fmt_vastr("unknown server-reply '%s'", mixi_srv_res), DEBUG);
	return -1;
  }
}



static int
mixi_handshake(void)
{
  int status;
  char buf[1024];
  CURL *curl;
  char *post_email, *post_password;

  curl = curl_easy_init();

  curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 1);

  curl_easy_setopt(curl, CURLOPT_URL, MIXI_LOGIN_URL);

  curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, mixi_store_res);

  memset(mixi_curl_errbuf, 0, sizeof(mixi_curl_errbuf));

  curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, mixi_curl_errbuf);
  curl_easy_setopt(curl, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);

  post_email = curl_escape(mixi_email, 0);
  post_password = curl_escape(mixi_password, 0);

  sprintf(buf, "email=%s&password=%s", post_email, post_password);
  curl_easy_setopt(curl, CURLOPT_POSTFIELDS, buf);

  curl_free(post_email);
  curl_free(post_password);

  pdebug("send session ID request.", DEBUG);

  status = curl_easy_perform(curl);
  curl_easy_cleanup(curl);

  if (status) {
	pdebug(mixi_curl_errbuf, DEBUG);
	mixi_hs_errors ++;
	mixi_free_res();
	return -1;
  }
  
  if (mixi_parse_hs_res() < 0) {
	mixi_hs_errors ++;
	mixi_free_res();
	return -1;
  }

  mixi_hs_errors = 0;
  mixi_hs_status = 1;

  mixi_free_res(); 

  pdebug("handshake OK", DEBUG);

  return 0;
}


static int
mixi_parse_sb_res(void)
{
  char *ptr1, *ptr2;

  if (!mixi_srv_res_size) {
	pdebug("No response from server", DEBUG);
	return -1;
  }

  *(mixi_srv_res + mixi_srv_res_size) = 0;

  if ((ptr1 = strstr(mixi_srv_res, "status="))) {
	ptr1 += 7;
	if ((ptr2 = (strchr(ptr1, '\n')))) *ptr2 = '\0'; 
  }

  if (strncmp(ptr1, "OK", 2) == 0) {
	pdebug("submission: OK", DEBUG);
	return 0;
  } else if (strncmp(ptr1, "BAD_SESSION", 11) == 0) {
	pdebug("submission: bad session ID", DEBUG);
	mixi_giveup = 0;
	mixi_hs_status = 0;
	if (mixi_session_id) free(mixi_session_id);
	mixi_session_id = NULL;
	if(mixi_upload_url != NULL) free(mixi_upload_url);
	mixi_upload_url = NULL;

	return -1;
  } else if (strncmp(ptr1, "BAD_DATA", 8) == 0) {
	pdebug("submission: bad data format", DEBUG);
	// don't retly
	return 0;
  } else {
	pdebug("submission: unknown error", DEBUG);
	// don't retly
	return 0;
  }
}


static gchar *
mixi_itemtag(const char *tag, char *str)
{
  static char *buf = NULL;
  if (buf != NULL) free(buf);

  buf = (char *)malloc(strlen(tag) * 2 + strlen(str) + 18 + 1);

  sprintf(buf, "%%3C%s%%3E%s%%3C%%2F%s%%3E%%0A", tag, str, tag);

  return buf;
}



static int
mixi_generateentry(GString *submission)
{
  int i;
  item_t *item;

  if (!mixi_q_len()) return 0;

  i = 0;

  mixi_q_peekall(1);

  while ((item = mixi_q_peekall(0))) {
	g_string_append(submission, "%3Ctrack%3E%0A");
	g_string_append(submission, mixi_itemtag("name", I_TITLE(item)));
	g_string_append(submission, mixi_itemtag("artist", I_ARTIST(item)));
	g_string_append(submission, mixi_itemtag("album", I_ALBUM(item)));
	g_string_append(submission, mixi_itemtag("genre", I_GENRE(item)));
	g_string_append(submission, mixi_itemtag("timestamp", I_TIME(item)));
	g_string_append(submission, "%3C%2Ftrack%3E%0A");
	i++;
  }

  return i;
}



static int
mixi_submitentry(gchar *entry)
{

  CURL *curl;
  struct curl_slist *slist = NULL;
  int status;
  GString *submission;
  char buf[1024];

  curl = curl_easy_init();
  curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 1);
  curl_easy_setopt(curl, CURLOPT_URL, mixi_upload_url);
  curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, mixi_store_res);

  curl_easy_setopt(curl, CURLOPT_USERAGENT, MIXI_USER_AGENT);

  pdebug(mixi_fmt_vastr("mixi_upload_url: %s", mixi_upload_url), DEBUG);


  sprintf(buf, "session_id=%s&data=%%3C%%3Fxml%%20version%%3D%%221%%2E0%%22%%20encoding%%3D%%22UTF%%2D8%%22%%3F%%3E%%0A", mixi_session_id);
  submission = g_string_new(buf);

  g_string_append(submission, "%3Creport%3E%0A");

  g_string_append(submission, entry);

  g_string_append(submission, "%3C%2Freport%3E%0A");

  curl_easy_setopt(curl, CURLOPT_POSTFIELDS, (char *)submission->str); 

  memset(mixi_curl_errbuf, 0, sizeof(mixi_curl_errbuf));

  curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, mixi_curl_errbuf); 
  curl_easy_setopt(curl, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);

  slist = curl_slist_append(slist, "Cache-Control: no-cache");
  slist = curl_slist_append(slist, "Accept:");

  curl_easy_setopt(curl, CURLOPT_HTTPHEADER, slist);
 
  status = curl_easy_perform(curl); 
  pdebug(mixi_fmt_vastr("%s\n", (char *)submission->str), DEBUG); 
  //status = 0;
  //status = -1; // for debug

  curl_slist_free_all(slist);

  curl_easy_cleanup(curl);

  g_string_free(submission,TRUE);

  if (status) { 
	pdebug(mixi_curl_errbuf, DEBUG); 
	mixi_sb_errors++;

	mixi_free_res();
	return -1;
  }

  if (mixi_parse_sb_res() < 0) {
	mixi_sb_errors++;
	mixi_free_res();  

	return -1;
  }
  mixi_free_res();
  return 0;
}


static void
mixi_handlequeue(pthread_mutex_t mutex)
{

  GString *submitentry;
  int nsubmit, i;
  int wait;

  if (mixi_submit_timeout < time(NULL)) {

	submitentry = g_string_new("");

	pthread_mutex_lock(&mutex);

	nsubmit = mixi_generateentry(submitentry);

	pthread_mutex_unlock(&mutex);

	if (nsubmit > 0) {
	  
	  if (! mixi_submitentry(submitentry->str)) {

		pthread_mutex_lock(&mutex); 

		for (i = 0; i < nsubmit; i ++) {
		  if (!mixi_q_get()) {
			pdebug("Fatal error: queue has broken", DEBUG);
			break;
		  }
		}

		dump_queue(); 

		pthread_mutex_unlock(&mutex); 
		mixi_sb_errors = 0;
	  }

	  if (mixi_sb_errors) {
		if (mixi_sb_errors < MIXI_RETRY_SHORT_COUNT) {
		  /* re-try after 1min */
		  wait = MIXI_RETRY_SHORT;
		} else if (mixi_sb_errors < MIXI_RETRY_LONG_COUNT) {
		  wait = MIXI_RETRY_LONG(mixi_sb_errors);
		} else {
		  wait = MIXI_RETRY_MAX;
		}
		mixi_submit_timeout = time(NULL) + wait;
		pdebug(mixi_fmt_vastr("re-try after %d[sec]", wait), DEBUG);
	  }
	}
	g_string_free(submitentry,TRUE);
  }
}


static char *
get_cache_file(void)
{
  char *home, *buf;

  buf = (char *)malloc(PATH_MAX);

  if (! (home = getenv("HOME"))) return NULL;

#ifdef MAKE_XMMS
  snprintf(buf, PATH_MAX, "%s/.xmms/mixi_music_queue.txt", home);
#endif
#ifdef MAKE_AUDACIOUS
  snprintf(buf, PATH_MAX, "%s/.audacious/mixi_music_queue.txt", home);
#endif

  buf[PATH_MAX - 1] = '\0';

  pdebug(mixi_fmt_vastr("cache_file %s", buf), DEBUG);

  return buf;
}


static void
read_cache(void)
{
  FILE *fd;
  char buf[PATH_MAX], *cache = NULL, *ptr1, *ptr2;
  char *cache_file;
  int cache_size, written;
  item_t *item;
  int err = 0;

  cache_size = 0;
  written = 0;
 
  cache_file = get_cache_file();
  strcpy(buf, cache_file);
  free(cache_file);

  if (!(fd = fopen(buf, "r"))) return;

  pdebug(mixi_fmt_vastr("Opening %s", buf), DEBUG);

  while (!feof(fd)) {
	cache_size += CACHE_SIZE;
	cache = realloc(cache, cache_size + 1);
	written += fread(cache + written, 1, CACHE_SIZE, fd);
	cache[written] = '\0';

	if (ferror(fd)) {
	  free(cache);
	  fclose(fd);
	  return;
	}
  }

  ptr1 = cache;

  while (ptr1 < cache + written - 1) {

	char *title = NULL, *artist = NULL, *len = NULL, 
	  *time = NULL, *album = NULL, *genre = NULL;

	pdebug("Pushed:", DEBUG);

	ptr2 = strchr(ptr1, ' ');
	if (!ptr2) {
	  err = 1;
	  goto READCACHE_CLEANUP;
	}
	artist = calloc(1, ptr2 - ptr1 + 1);
	strncpy(artist, ptr1, ptr2 - ptr1);


	ptr1 = ptr2 + 1;

	ptr2 = strchr(ptr1, ' ');
	if (!ptr2) {
	  err = 1;
	  goto READCACHE_CLEANUP;
	}
	title = calloc(1, ptr2 - ptr1 + 1);
	strncpy(title, ptr1, ptr2 - ptr1);

	ptr1 = ptr2 + 1;

	ptr2 = strchr(ptr1, ' ');
	if (!ptr2) {
	  err = 1;
	  goto READCACHE_CLEANUP;
	}
	len = calloc(1, ptr2 - ptr1 + 1);
	strncpy(len, ptr1, ptr2 - ptr1);

	ptr1 = ptr2 + 1;

	ptr2 = strchr(ptr1, ' ');
	if (!ptr2) {
	  err = 1;
	  goto READCACHE_CLEANUP;
	}
	time = calloc(1, ptr2 - ptr1 + 1);
	strncpy(time, ptr1, ptr2 - ptr1);

	ptr1 = ptr2 + 1;

	ptr2 = strchr(ptr1, ' ');
	if (!ptr2) {
	  err = 1;
	  goto READCACHE_CLEANUP;
	}
	album = calloc(1, ptr2 - ptr1 + 1);
	strncpy(album, ptr1, ptr2 - ptr1);

	ptr1 = ptr2 + 1;

	ptr2 = strchr(ptr1, '\n');

	if (!ptr2) {
	  err = 1;
	  goto READCACHE_CLEANUP;
	}

	*ptr2 = '\0';
	genre = calloc(1, strlen(ptr1) + 1);
	strncpy(genre, ptr1, strlen(ptr1));

	*ptr2 = '\n';

	ptr1 = ptr2 + 1;

	item = mixi_q_put2(artist, title, len, time, album, genre);

  READCACHE_CLEANUP:
	if (artist) free(artist);
	if (title) free(title);
	if (len) free(len);
	if (time) free(time);
	if (album) free(album);
	if (genre) free(genre);

	if (err) break;
  }

  pdebug("Done loading cache.", DEBUG);
}


static void
dump_queue(void)
{
  FILE *fd;
  item_t *item;
  char buf[PATH_MAX];
  char *cache_file;

  cache_file = get_cache_file();
  strcpy(buf, cache_file);
  pdebug(mixi_fmt_vastr("Opening %s", cache_file), DEBUG);
  free(cache_file);

  pdebug(mixi_fmt_vastr("dump_queue: %s", buf), DEBUG);

  if (!(fd = fopen(buf, "w"))){
	pdebug("failed: fopen", DEBUG);
	return;
  }

  pdebug(mixi_fmt_vastr("Opening %s", buf), DEBUG);

  mixi_q_peekall(1);

  while ((item = mixi_q_peekall(0))) { 

	pdebug(mixi_fmt_vastr("title: %s", I_TITLE(item)), DEBUG);

	fprintf(fd, "%s %s %s %s %s %s\n",
			I_ARTIST(item),
			I_TITLE(item),
			I_LEN(item),
			I_TIME(item),
			I_ALBUM(item),
			I_GENRE(item));
  }

  fclose(fd);
}

/* This was made public */

void mixi_cleaner(void)
{
  dump_queue();
  mixi_q_free();
  pdebug("mixi music shutting down", DEBUG);
}

static void
mixi_checkhandshake(void)
{
  int wait;

  if (mixi_hs_status) return;

  if (mixi_hs_timeout < time(NULL)) {

	mixi_handshake();

	if (mixi_hs_errors) {
	  if (mixi_hs_errors < MIXI_RETRY_SHORT_COUNT)
		/* re-try after 1min */
		wait = MIXI_RETRY_SHORT;
	  else if (mixi_hs_errors < MIXI_RETRY_LONG_COUNT)
		wait = MIXI_RETRY_LONG(mixi_hs_errors - MIXI_RETRY_SHORT_COUNT);
	  else
		wait = MIXI_RETRY_MAX;

	  mixi_hs_timeout = time(NULL) + wait;

	  pdebug(mixi_fmt_vastr("re-try after %d[sec]", wait), DEBUG);
	}
  }

}



/**** Public *****/

/* Called at session startup*/

void
mixi_music_init(char *email, char *password)
{

  mixi_hs_status = 0;
  mixi_hs_timeout = 0;
  mixi_hs_errors = 0;
  mixi_sb_errors = 0;
  mixi_submit_timeout = 0;
  mixi_srv_res_size = 0;
  mixi_giveup = 0;
  mixi_major_error_present = 0;
  mixi_bad_users = 0;


  mixi_upload_url = NULL;
  mixi_srv_res = NULL;
  mixi_major_error = NULL;

  mixi_email = strdup(email);
  mixi_password = strdup(password);

  read_cache();

  pdebug("mixi music starting up", DEBUG);
}

void mixi_addentry(pthread_mutex_t mutex, metatag_t *meta, int len)
{
  pthread_mutex_lock(&mutex);

  mixi_q_put(meta, len); 

  /*
   * This will help make sure the queue will be saved on a nasty
   * segfault...
   */
  dump_queue(); 

  pthread_mutex_unlock(&mutex);
}

/* Call periodically from the plugin */
int
mixi_idle(pthread_mutex_t mutex)
{
  mixi_checkhandshake();

  if (mixi_hs_status) mixi_handlequeue(mutex); 

  return mixi_giveup;
}
