#ifdef MAKE_XMMS
# include <xmms/plugin.h>
# include <xmms/configfile.h>
# include <xmms/xmmsctrl.h>
#endif

#ifdef MAKE_AUDACIOUS
# include <audacious/plugin.h>
# include <audacious/configdb.h>
# include <audacious/beepctrl.h>
# define ConfigFile ConfigDb
# define xmms_cfg_open_default_file bmp_cfg_db_open
# define xmms_cfg_read_string bmp_cfg_db_get_string
# define xmms_cfg_free bmp_cfg_db_close
#endif

#include <pthread.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <wchar.h>
#include <sys/time.h>

#include "tags/include/tags.h"
#include "mixi_music.h"
#include "gtkstuff.h"
#include "config.h"
#include "fmt.h"
#include "tags/include/unicode.h"

#define XM_CS xmms_mixi_music.xmms_session

static void init(void);
static void cleanup(void);
static void *mm_thread(void *);
static void *hs_thread(void *);
static int going;

static pthread_t pt_mixi_music;
static pthread_mutex_t m_mixi_music = PTHREAD_MUTEX_INITIALIZER;
static pthread_t pt_handshake;

static GeneralPlugin xmms_mixi_music =
{
  NULL,
  NULL,
  -1,
  NULL,
  init,
  mixi_about_show,
  mixi_configure_dialog,
  cleanup
};


static void
init(void)
{
  char *email = NULL, *password = NULL;

  ConfigFile *cfgfile;

  going = 1;

  if ((cfgfile = xmms_cfg_open_default_file()) != NULL) {
	xmms_cfg_read_string(cfgfile, "miximusic", "email",
						 &email);
	xmms_cfg_read_string(cfgfile, "miximusic", "password",
						 &password);
	xmms_cfg_free(cfgfile);
  }

  if ((!email || !password) || (!*email || !*password)) {
	pdebug("email/password not found - not starting", DEBUG);
	going = 0;
	return;
  }

  mixi_music_init(email, password);

  g_free(email);
  g_free(password);

  if (pthread_create(&pt_mixi_music, NULL, mm_thread, NULL)) {
	pdebug("Error creating mixi music thread", DEBUG);
	going = 0;
	return;
  }

  mixi_init_errorbox_done();

  if (pthread_create(&pt_handshake, NULL, hs_thread, NULL)) {
	pdebug("Error creating handshake thread", DEBUG);
	going = 0;
	return;
  }

  pdebug("plugin started", DEBUG);
}


static void
cleanup(void)
{
  void *dummy;

  if (!going) return;

  pdebug("about to lock mutex", DEBUG);
  pthread_mutex_lock(&m_mixi_music);

  pdebug("locked mutex", DEBUG);

  going = 0;

  pthread_mutex_unlock(&m_mixi_music);

  pdebug("joining threads", DEBUG);

  pthread_join(pt_mixi_music, &dummy); 

  pthread_cancel(pt_handshake);

  pthread_join(pt_handshake, &dummy); 

  mixi_cleaner();
}


static char
ishttp(const char *a)
{
  char *tmp, *bp;
  int status = 0;

  if (!a || !*a) return 0;

  tmp = strdup(a);

  for (bp = tmp; *bp; bp++)
	*bp = toupper(*bp);

  if (strstr(tmp, "HTTP://")) status = -1;

  free(tmp);

  return status;
}



static void *
mm_thread(void *data)
{
  int run = 1;
  char *file_c = NULL, *file_p = NULL; 

  enum playstatus {
	ps_stop, ps_play, ps_pause
  } ps_c, ps_p = ps_stop;

  int submitted = 0;

  enum state {
	start, stop, pause, restart, playing, pausing, stopping
  } playstate;

  int changed;
  int seeked;
  int dosubmit = 0;
  int repeat;
  int filechanged;
  int rewind; 
  char *charpos;
  int len = 0;

  int playtime_c, playtime_p = 0, playtime_i = 0, playtime_ofs = 0;
  long long int time_c, time_i = 0;
  int  pos_p = 0, pos_c;
  int  playlistlen_p = 0, playlistlen_c;

  struct timeval timetmp;
  
  while (run) {
	// Error catching
	if (mixi_catch_error()) {
	  mixi_errorbox_show(mixi_fetch_error());
	  if (mixi_get_errorbox_done()) {
		mixi_init_errorbox_done();
		mixi_clear_error();
	  }
	}

	// current music number
	pos_c = xmms_remote_get_playlist_pos(XM_CS);
	// current file name
	file_c = xmms_remote_get_playlist_file(XM_CS, pos_c); 
	// total number
	playlistlen_c = xmms_remote_get_playlist_length(XM_CS);
	// current playtime
	playtime_c = xmms_remote_get_output_time(XM_CS); 
	// total length
	len = xmms_remote_get_playlist_time(XM_CS, pos_c); 

	// current time (ms)
	gettimeofday(&timetmp, NULL);
	time_c = timetmp.tv_sec * 1000 + timetmp.tv_usec / 1000; 

	// current status
	if (xmms_remote_is_paused(XM_CS))
	  ps_c = ps_pause;
	else if (xmms_remote_is_playing(XM_CS))
	  ps_c = ps_play;
	else
	  ps_c = ps_stop;

	// repeat setting
	repeat = xmms_remote_is_repeat(XM_CS);

	if (ps_p == ps_stop && ps_c == ps_stop)        playstate = stopping;
	else if (ps_p == ps_stop && ps_c == ps_play)   playstate = start;
	// else if (ps_p == ps_stop && ps_c == ps_pause) ;
	else if (ps_p == ps_play && ps_c == ps_play)   playstate = playing;
	else if (ps_p == ps_play && ps_c == ps_stop)   playstate = stop;
	else if (ps_p == ps_play && ps_c == ps_pause)  playstate = pause;
	else if (ps_p == ps_pause && ps_c == ps_pause) playstate = pausing;
	else if (ps_p == ps_pause && ps_c == ps_play)  playstate = restart;
	else if (ps_p == ps_pause && ps_c == ps_stop)  playstate = stop;
	else playstate = stopping;

	// filename has changed
	if (!(file_p == NULL && file_c == NULL) &&
	    ((file_p == NULL && file_c != NULL) ||
		 (file_p != NULL && file_c == NULL) ||
		 (file_p != NULL && file_c != NULL && strcmp(file_c, file_p)))) {
	  filechanged = 1;
	  pdebug("*** filechange ***", DEBUG);
	} else {
	  filechanged = 0;
	}

	if (file_c == NULL) len = 0;

	// whole rewind has occurred (maybe)
	if (len != 0 && len - (playtime_p - playtime_c) < 3000) {
	  rewind = 1;
	  pdebug("*** rewind ***", DEBUG);
	} else {
	  rewind = 0;
	}

	changed = 0;
	seeked = 0;

	switch (playstate) {
	case start:
	  pdebug("*** START ***", DEBUG);
	  break;
	case stop:
	  pdebug("*** STOP ***", DEBUG);
	  len = 0;
	  break;
	case pause: 
	  pdebug("*** PAUSE ***", DEBUG);
	  playtime_ofs += playtime_c - playtime_i; // save playtime
	  break;
	case restart: 
	  pdebug("*** RESTART ***", DEBUG);
	  playtime_i  = playtime_c; // restore playtime
	  break;
	case playing:
	  if ((playtime_c < playtime_p) || // back
		  ((playtime_c - playtime_i) - (time_c - time_i) > 3000) //forward
		  ) 
		seeked = 1;

	  if (filechanged || // filename has changed
		  (!filechanged && // filename has not changed...
			//(( rewind && (repeat && (!advance || (pos_c == 0 && playlistlen_c == 1 )))) || 
			// looping with only one file
		   ((pos_c == 0 && playlistlen_c == 1 && repeat && rewind) || 
			 // looping?
			(pos_p == pos_c && rewind) ||
			(pos_p != pos_c && seeked) ||
			// skip from current music to next music, 
			//  which has the same filename as previous one
			(pos_p < pos_c && playtime_c < playtime_p) || 
			// current song has removed from playlist 
			// but the next (following) song has the same filename
			(playlistlen_p > playlistlen_c && playtime_c < playtime_p)))) {

		pdebug("*** CHANGE ***",DEBUG);
		pdebug(mixi_fmt_vastr(" filechanged = %d",filechanged),DEBUG);
		pdebug(mixi_fmt_vastr(" pos_c = %d",pos_c),DEBUG);
		pdebug(mixi_fmt_vastr(" pos_p = %d",pos_p),DEBUG);
		pdebug(mixi_fmt_vastr(" rewind = %d", rewind),DEBUG);
		pdebug(mixi_fmt_vastr(" seeked = %d", seeked),DEBUG);
		pdebug(mixi_fmt_vastr(" playtime_c = %d", playtime_c),DEBUG);
		pdebug(mixi_fmt_vastr(" playtime_p = %d", playtime_p),DEBUG);
		pdebug(mixi_fmt_vastr(" playlistlen_c = %d", playlistlen_p),DEBUG);
		pdebug(mixi_fmt_vastr(" playlistlen_p = %d", playlistlen_p),DEBUG);
		changed = 1;
		seeked = 0;
	  } else if (seeked) { 
		seeked = 1;
		pdebug("*** SEEK ***", DEBUG);
	  }
	  break;
	case pausing: 
	  if (playtime_c != playtime_p) {
		pdebug("*** SEEK ***", DEBUG);
		seeked = 1;
	  }
	  break;
	case stopping:
	  len = 0;
	  break;
	default:
	  pdebug("*** unknown state tranfer!!! ***", DEBUG);
	  break;
	}

	if (playstate == start || changed || (seeked && !submitted)) {
	  // reset counter
	  pdebug(" <<< reset counter >>>", DEBUG);

	  submitted = 0;
	  playtime_ofs = 0;
	  playtime_i = playtime_c;
	  time_i = time_c;

	} else {
	  // check playtime for submitting
	  if (!submitted) {
		if (len > 30 * 1000 &&
			//len < 30 *60 * 1000 &&  // crazy rule!!!
			((playtime_ofs + playtime_c - playtime_i > len / 2) ||
			 (playtime_ofs + playtime_c - playtime_i > 240 * 1000) 
			 //(playtime_c - playtime_i > 10 * 1000)  // for debug
			 )) {
		  pdebug("*** submitting requirements are satisfied.",DEBUG);
		  pdebug(mixi_fmt_vastr("    len = %d, playtime = %d",
						   len / 1000, (playtime_c - playtime_i)/1000 ), DEBUG);
		  submitted = 1;
		  dosubmit = 1;
		}
	  }
	}

	// keep current value for next iteration
	ps_p = ps_c;
	file_p = file_c;
	playtime_p = playtime_c;
	pos_p = pos_c;
	playlistlen_p = playlistlen_c;

	if (dosubmit) {
	  char *fname, *tmp = NULL;
	  int track = 0;
	  metatag_t *meta;

	  pdebug("*** do submit", DEBUG);

	  dosubmit = 0;

	  meta = mixi_metatag_new();

	  fname = xmms_remote_get_playlist_file(0, pos_c);

	  /* skip file on network */
	  if (ishttp(fname)) {
		g_free(fname);
		continue;
	  }

	  charpos = strrchr(fname, '.');

	  if (charpos != NULL && !mixi_fmt_strncasecmp(charpos + 1, "cda", 3)) {
		ConfigFile *cfgfile;
		  
		if ((cfgfile = xmms_cfg_open_default_file()) != NULL) {

		  tmp = strrchr(fname, '.');
		  *tmp = '\0';
		  track = (char)atoi(tmp - 2);
		  pdebug(mixi_fmt_vastr("Track: %d", track), DEBUG);
		  xmms_cfg_read_string(cfgfile, "CDDA", "device", &fname);
		  xmms_cfg_free(cfgfile);
		  pdebug(mixi_fmt_vastr("CD Device: %s", fname), DEBUG);
		}
	  }

	  mixi_get_tag_data(meta, fname, track); // parse music file directly...

	  if (meta->artist && meta->title)
		mixi_addentry(m_mixi_music, meta, len/1000); 
	  else
		pdebug("couldn't determine artist - "
			   "title, not submitting",
			   DEBUG);

	  g_free(tmp);
	  g_free(fname);
	  mixi_metatag_delete(meta);
	}

	pthread_mutex_lock(&m_mixi_music);
	run = going;
	pthread_mutex_unlock(&m_mixi_music);
	usleep(100000);
  }
  pdebug("mixi music thread: exiting", DEBUG);
  pthread_exit(NULL);
}


static void *
hs_thread(void *data)
{
  int run = 1;
	
  while (run) {
	if (mixi_idle(m_mixi_music)) {
	  pdebug("Giving up due to fatal error", DEBUG);
	  pthread_mutex_lock(&m_mixi_music);
	  going = 0;
	  pthread_mutex_lock(&m_mixi_music);
	}
	pthread_mutex_lock(&m_mixi_music);
	run = going;
	pthread_mutex_unlock(&m_mixi_music);
	sleep(1);
  }

  pdebug("handshake thread: exiting", DEBUG);
  pthread_exit(NULL);
}

GeneralPlugin *
get_gplugin_info(void)
{
#ifdef MAKE_XMMS
  xmms_mixi_music.description = g_strdup_printf("XMMS-mixi-Music %s",
											   VERSION);
#endif
#ifdef MAKE_AUDACIOUS
  xmms_mixi_music.description = g_strdup_printf("Audacious-mixi-Music %s",
											   VERSION);
#endif
  return &xmms_mixi_music;
}
